@@ -35,30 +35,13 @@ class SimProtocol(Protocol):
     OFFSETS = [0x10000]
 
     def __init__(self, url, log):
-        if callable(url):
-            self._plcfunc = url
-        else:
-            try:
-                if not url.startswith('sim://'):
-                    raise Exception('URL should start with sim://')
-                with open(url[6:]) as fp:
-                    code = fp.read()
-                globs = {}
-                # TODO: execute with the correct filename to get
-                # filename  and code in tracebacks
-                exec(code, globs)  # pylint: disable=exec-used
-                self._plcfunc = globs['Main']
-            except Exception as err:
-                raise ApiError('invalid sim address, must be '
-                               'sim:///path/to/file.py '
-                               'with a Main() function') from err
-
+        self._plcfunc = url
         self._cond = threading.Condition()
         self._thread = None
         self._stopflag = False
         self._exception = None
 
-        Protocol.__init__(self, url, log)
+        super().__init__(url, log)
 
     def connect(self):
         if self._thread and self._thread.is_alive():
@@ -67,6 +50,11 @@ class SimProtocol(Protocol):
         self._thread = threading.Thread(target=self._plc_thread, daemon=True,
                                         name=f'SimPLC {self.url}')
         self._thread.start()
+        # wait for one cycle, and raise if an exception occurred
+        with self._cond:
+            self._cond.wait()
+        if self._exception:
+            raise self._exception
         self._connect_callback(True)
         self.connected = True
 
@@ -81,7 +69,7 @@ class SimProtocol(Protocol):
         with self._cond:
             self._cond.wait()
             try:
-                return runtime.mem.read(self.offset + addr, length)
+                return self.my_mem.read(self.offset + addr, length)
             except RuntimeError as err:
                 raise CommError(f'reading {addr}({length}): {err}') from err
 
@@ -89,11 +77,29 @@ class SimProtocol(Protocol):
         with self._cond:
             self._cond.wait()
             try:
-                runtime.mem.write(self.offset + addr, data)
+                self.my_mem.write(self.offset + addr, data)
             except RuntimeError as err:
                 raise CommError(f'writing {addr}({len(data)}): {err}') from err
 
     def _plc_thread(self):
+        self.my_mem = runtime.Memory()
+        try:
+            if not self._plcfunc.startswith('sim://'):
+                raise Exception('URL should start with sim://')
+            with open(self._plcfunc[6:], encoding='utf-8') as fp:
+                code = fp.read()
+            globs = {}
+            # TODO: execute with the correct filename to get
+            # filename  and code in tracebacks
+            exec(code, globs)  # pylint: disable=exec-used
+            self._plcfunc = globs['Main']
+        except Exception as err:
+            self._exception = ApiError(
+                'invalid sim address, must be '
+                'sim:///path/to/file.py '
+                'with a Main() function')
+            self.log.error('%s, aborting simulation', self._exception)
+            self._exception.__cause__ = err
         while not self._stopflag:
             with self._cond:
                 try:
