@@ -18,6 +18,7 @@
 #
 # Module authors:
 #   Georg Brandl <g.brandl@fz-juelich.de>
+#   Enrico Faulhaber <enrico.faulhaber@frm2.tum.de>
 #
 # *****************************************************************************
 
@@ -25,11 +26,13 @@
 
 from collections import namedtuple
 
-from zapf import spec
-from zapf.device import Device
+from zapf import SpecError, spec
+from zapf.device import TYPECODE_MAP, Device, FMT_LIMITS
 from zapf.io import PlcIO
+from zapf.spec import FMT_TO_BASETYPE, v_2015_02
+from zapf.spec.v_2021_09 import fix_unit
 
-DeviceInfo = namedtuple('DeviceInfo', 'number name addr typecode info')
+DeviceInfo = namedtuple('DeviceInfo', 'name addr typecode info')
 
 
 class Scanner:
@@ -65,6 +68,9 @@ class Scanner:
 
         Example: ``tango://192.168.201.2:10000/box/plc/ads``
     """
+
+    IGNORE_BAD_DEVICES = True
+
     def __init__(self, io_or_proto, log):
         self.log = log
         if isinstance(io_or_proto, PlcIO):
@@ -123,7 +129,7 @@ class Scanner:
         * ``absmin`` and ``absmax``: float, the limits of the main value
         * ``params``: a list of parameter names if the device has params
         * ``funcs``: a list of special function names if the device has such
-        * ``aux_strings``: a list of 24 strings containing the interpretation
+        * ``aux``: a specification of bitfields/flags containing the interpretation
           of the up to 24 freely definable bits in the device status.
           Strings that are not defined are empty.
 
@@ -136,8 +142,8 @@ class Scanner:
         # check which method we need to scan
         method = getattr(self, '_scan_' + self.io.indexer.magicstr, None)
         if not method:
-            raise RuntimeError('Magic %s is supported, but no scanner method '
-                               'available' % self.io.indexer.magicstr)
+            raise RuntimeError(f'Magic {self.io.indexer.magicstr} is '
+                               'supported, but no scanner method available')
 
         yield from method()
 
@@ -154,8 +160,8 @@ class Scanner:
                              devinfo.typecode)
             return None
         else:
-            return devcls(devinfo.number, devinfo.name, devinfo.addr,
-                          devinfo.typecode, devinfo.info, self.io, self.log)
+            return devcls(devinfo.name, devinfo.addr, devinfo.typecode,
+                          devinfo.info, self.io, self.log)
 
     def get_devices(self):
         """Scan the PLC and yield device objects.
@@ -178,15 +184,15 @@ class Scanner:
             if typecode == 0:
                 break
 
-            # if there is not valid data in the infostruct, query individually
-            if size + addr + flags == 0 and not name:
-                size = indexer.query_word(devnum, spec.INFO_SIZE)
+            # if there is no valid data in the infostruct, query individually
+            if size + addr + flags == 0:
+                size = indexer.query_word(devnum, spec.v_2015_02.INFO_SIZE)
                 if not size:
-                    size = 2 * (typecode & 0xff)
-                addr = indexer.query_word(devnum, spec.INFO_ADDR)
+                    size = 2 * (typecode & 0xFF)
+                addr = indexer.query_word(devnum, spec.v_2015_02.INFO_ADDR)
                 if not addr:
                     addr = next_addr
-                unit = indexer.query_unit(devnum, spec.INFO_UNIT)
+                unit = indexer.query_unit(devnum, spec.v_2015_02.INFO_UNIT)
                 flags = None
                 lowlevel = False
                 absmin = -spec.FLOAT32_MAX
@@ -197,7 +203,7 @@ class Scanner:
             # this might be empty even if a valid infostruct was present,
             # e.g. if the full name didn't fit behind the 20 previous bytes
             if not name:
-                name = indexer.query_string(devnum, spec.INFO_NAME)
+                name = indexer.query_string(devnum, spec.v_2015_02.INFO_NAME)
 
             # extract info about parameters and special functions
             devclass = typecode >> 13
@@ -205,39 +211,144 @@ class Scanner:
             functions = []
             if devclass in (1, 2):
                 if devclass == 1:  # FlatDevices: a list of indices
-                    plist = indexer.query_bytes(devnum, spec.INFO_PARAMS)
+                    plist = indexer.query_bytes(devnum, spec.v_2015_02.INFO_PARAMS)
                     param_ids = [p for p in plist if p]
                 else:  # ParamDevices: a bitmap of indices
-                    param_ids = indexer.query_bitmap(devnum, spec.INFO_PARAMS)
+                    param_ids = indexer.query_bitmap(devnum,
+                                                     spec.v_2015_02.INFO_PARAMS)
                 param_ids.sort()  # ... which *should* be unnecessary
                 for p in param_ids:
-                    if spec.is_function(p):
-                        functions.append(spec.Parameters[p])
+                    if spec.v_2015_02.is_function(p):
+                        functions.append(p)
                     else:
-                        parameters.append(spec.Parameters[p])
+                        parameters.append(p)
 
             # extract AUX string labels
             aux_strings = [''] * 24
             if flags is None:  # old way, stop reading the first empty one
                 for idx in range(24):
-                    aux_string = indexer.query_string(devnum,
-                                                      spec.INFO_AUX1 + idx)
+                    aux_string = indexer.query_string(
+                        devnum, spec.v_2015_02.INFO_AUX1 + idx
+                    )
                     aux_strings[idx] = aux_string.strip()
                     if not aux_string:
                         break
             else:  # new way, only read the relevant (flagged ones)
                 for idx in range(24):
                     if flags & (1 << idx):
-                        aux_string = indexer.query_string(devnum,
-                                                          spec.INFO_AUX1 + idx)
+                        aux_string = indexer.query_string(
+                            devnum, spec.v_2015_02.INFO_AUX1 + idx
+                        )
                         aux_strings[idx] = aux_string.strip()
 
-            info = dict(lowlevel=lowlevel, unit=unit,
+            # XXX: convert to new style! (see _scan_2021_09)
+            flags = ['lowlevel'] if lowlevel else []
+            aux = []
+            for i, s in enumerate(aux_strings):
+                if s:
+                    # convert to Flag, bitfields dont exist in 2015_02
+                    aux.append((i, s))
+
+            try:
+                tce = TYPECODE_MAP[typecode]
+            except KeyError:
+                raise SpecError(f'device {name!r} has unsupported typecode '
+                                f'{typecode:#06x}') from None
+
+            params = {}  # derive from parameters
+            funcs = {}  # derive from functions
+            if tce.num_params:
+                if tce.has_pctrl:
+                    # convert params and funcs
+                    for p in parameters:
+                        par = {'idx': p}
+                        par.update(v_2015_02.PARAMETER_Specs[p])
+                        if 'unit' in par:
+                            par['unit'] = fix_unit(par.get('unit', ''), unit)
+                        params[par.pop('name')] = par
+                    for p in functions:
+                        par = {'idx': p}
+                        par.update(v_2015_02.FUNC_Specs[p])
+                        if par.get('argument', None):
+                            arg = {}
+                            arg.update(par['argument'])
+                            if 'unit' in arg:
+                                arg['unit'] = fix_unit(arg.get('unit', ''), unit)
+                            par['argument'] = arg
+                        funcs[par.pop('name')] = par
+                else:
+                    # convert flat params
+                    for i, p in enumerate(parameters):
+                        par = {'idx': i}  # parameter slot id
+                        par.update(v_2015_02.PARAMETER_Specs[p])
+                        if 'unit' in par:
+                            par['unit'] = fix_unit(par.get('unit', ''), unit)
+                        params[par.pop('name')] = par
+
+            basetype, width = FMT_TO_BASETYPE.get(tce.value_fmt, (None, 0))
+            # Keyword/RealValue are special
+            access = 'rw' if (tce.has_target or typecode >> 8 in (0x14, 0x15)) \
+                else 'ro'
+
+            # fixup limits, if needed
+            limits = FMT_LIMITS.get(tce.value_fmt)
+            if limits:
+                absmin = max(absmin, limits[0])
+                absmax = min(absmax, limits[1])
+
+            info = dict(unit=unit,
                         absmin=absmin, absmax=absmax,
-                        params=parameters, funcs=functions,
-                        aux_strings=aux_strings)
+                        params=params, funcs=funcs,
+                        aux=aux, tables={}, flags=flags,
+                        errid=[(0, 16, 'errid', {})],
+                        description='',
+                        access=access,
+                        basetype=basetype,
+                        width=width,
+                        enum_r=None,
+                        enum_w=None,
+                        )
             self.log.info('found device %s at addr %s with type %#x',
                           name, addr, typecode)
-            yield DeviceInfo(devnum, name, addr, typecode, info)
+            yield DeviceInfo(name, addr, typecode, info)
 
             next_addr = addr + size
+
+    def _scan_2021_09(self):
+        devid = self.io.indexer.plc_descriptor.last_device
+        seen_dev_names = set()
+
+        while devid:
+            try:
+                dev = self.io.indexer.query_descriptor(devid)
+            except SpecError as err:
+                self.log.error(err)
+                # try to re-read content without resolving
+                dev = self.io.indexer.query_descriptor_uncached(devid)
+                devid = dev.prev
+                continue
+
+            if dev.name in seen_dev_names:
+                self.log.debug(f'skipping device {dev.name!r}: a device with '
+                               'that name already exists.')
+                devid = dev.prev
+                continue
+            seen_dev_names.add(dev.name)
+
+            if dev.typecode not in TYPECODE_MAP:
+                self.log.error(f'device {dev.name!r} at address {dev.address:#06x} '
+                               f'has unsupported typecode {dev.typecode:#06x}')
+            try:
+                info = dev.get_info()
+            except SpecError as err:
+                self.log.error(f'device {dev.name!r}: {err}')
+                devid = dev.prev
+                if self.IGNORE_BAD_DEVICES:
+                    continue
+                raise SpecError(f'Bad Device {dev.name!r}: {err}') from err
+
+            self.log.info(f'found device {dev.name!r} at addr '
+                          f'{dev.address:#06x} with type {dev.typecode:#06x}')
+            yield DeviceInfo(info.pop('name'), info.pop('address'),
+                             info.pop('typecode'), info)
+            devid = dev.prev
