@@ -22,7 +22,7 @@
 #
 # *****************************************************************************
 
-from zapf.simulator.runtime import mem
+from zapf.simulator.runtime import local
 
 
 def adr(val):
@@ -34,9 +34,9 @@ def sizeof(val):
 
 
 def memset(addr, byte, count):
-    mem.write(addr, chr(byte).encode() * count)
+    local.mem.write(addr, chr(byte).encode() * count)
 
 
 def memcpy(toaddr, fromaddr, count):
-    data = mem.read(fromaddr, count)
-    mem.write(toaddr, data)
+    data = local.mem.read(fromaddr, count)
+    local.mem.write(toaddr, data)
