@@ -30,56 +30,121 @@ class NumProxy:
             other = other.value
         return self.value + other
 
+    def __radd__(self, other):
+        if isinstance(other, NumProxy):
+            other = other.value
+        return other + self.value
+
     def __sub__(self, other):
         if isinstance(other, NumProxy):
             other = other.value
         return self.value - other
 
+    def __rsub__(self, other):
+        if isinstance(other, NumProxy):
+            other = other.value
+        return other - self.value
+
     def __mul__(self, other):
         if isinstance(other, NumProxy):
             other = other.value
         return self.value * other
 
+    def __rmul__(self, other):
+        if isinstance(other, NumProxy):
+            other = other.value
+        return other * self.value
+
     def __floordiv__(self, other):
         if isinstance(other, NumProxy):
             other = other.value
         return self.value // other
 
+    def __rfloordiv__(self, other):
+        if isinstance(other, NumProxy):
+            other = other.value
+        return other // self.value
+
     def __truediv__(self, other):
         if isinstance(other, NumProxy):
             other = other.value
         return self.value / other
 
+    def __rtruediv__(self, other):
+        if isinstance(other, NumProxy):
+            other = other.value
+        return other / self.value
+
     def __mod__(self, other):
         if isinstance(other, NumProxy):
             other = other.value
         return self.value % other
 
+    def __rmod__(self, other):
+        if isinstance(other, NumProxy):
+            other = other.value
+        return other % self.value
+
+    def __divmod__(self, other):
+        if isinstance(other, NumProxy):
+            other = other.value
+        return divmod(self.value, other)
+
+    def __rdivmod__(self, other):
+        if isinstance(other, NumProxy):
+            other = other.value
+        return divmod(other, self.value)
+
     def __and__(self, other):
         if isinstance(other, NumProxy):
             other = other.value
         return self.value & other
 
+    def __rand__(self, other):
+        if isinstance(other, NumProxy):
+            other = other.value
+        return other & self.value
+
     def __or__(self, other):
         if isinstance(other, NumProxy):
             other = other.value
         return self.value | other
 
+    def __ror__(self, other):
+        if isinstance(other, NumProxy):
+            other = other.value
+        return other | self.value
+
     def __xor__(self, other):
         if isinstance(other, NumProxy):
             other = other.value
         return self.value ^ other
 
+    def __rxor__(self, other):
+        if isinstance(other, NumProxy):
+            other = other.value
+        return other ^ self.value
+
     def __rshift__(self, other):
         if isinstance(other, NumProxy):
             other = other.value
         return self.value >> other
 
+    def __rrshift__(self, other):
+        if isinstance(other, NumProxy):
+            other = other.value
+        return other >> self.value
+
     def __lshift__(self, other):
         if isinstance(other, NumProxy):
             other = other.value
         return self.value << other
 
+    def __rlshift__(self, other):
+        if isinstance(other, NumProxy):
+            other = other.value
+        return other << self.value
+
     def __lt__(self, other):
         if isinstance(other, NumProxy):
             other = other.value
