@@ -18,6 +18,7 @@
 #
 # Module authors:
 #   Georg Brandl <g.brandl@fz-juelich.de>
+#   Enrico Faulhaber <enrico.faulhaber@frm2.tum.de>
 #
 # *****************************************************************************
 
@@ -96,35 +97,9 @@ class UncasedMap(dict):
         self._keys = []
         for k, v in itertools.chain(kwds.items(), args):
             self[k] = v
+            self[k.lower()] = v
             self[v] = k
             self._keys.append(k)
-            try:
-                self[k.lower()] = v
-            except Exception:
-                pass
 
     def __getattr__(self, k):
         return self[k]
-
-    def init_keys(self):
-        """Return those possible keys that were keys in the initial mapping."""
-        return self._keys
-
-    def _gendefaults(self, prefix, idxlist):
-        for idx in idxlist:
-            if idx not in self:
-                k = '%s%d' % (prefix, idx)
-                self[k] = idx
-                self[idx] = k
-                self[k.lower()] = idx
-                self._keys.append(k)
-
-
-class ParamMap(UncasedMap):
-    def __init__(self, *args, **kwds):
-        UncasedMap.__init__(self, *args, **kwds)
-        self._gendefaults('Param', range(1, 128))
-        self._gendefaults('Command', range(128, 192))
-        self._gendefaults('Param', range(192, 224))
-        self._gendefaults('Command', range(224, 240))
-        self._gendefaults('Reserved', range(240, 256))
